import click
from .database import db_query
import pprint
import csv
# Grab all 19506 Data


def grab_hop_count(uuid):
    # grab the output of 10287 - Trace Route
    hop_count_data = db_query("select output from vulns where asset_uuid='{}' and plugin_id='10287';".format(uuid))

    # Send the raw data back
    return hop_count_data


@click.command(help="Evaluate Scan times")
def evaluate():

    # Pull all 19506 Plugins from the DB
    plugin_data = db_query("select asset_uuid, output from vulns where plugin_id='19506';")

    # Set some dicts for organizing Data
    scan_policy_dict = {}
    scanner_dict = {}
    scan_name_dict = {}

    # Open a CSV for export
    with open('evaluate.csv', mode='w') as csv_file:
        agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

        header_list = ["asset uuid", "Scan Name", "Scan Policy", "Scanner IP", "Scan Time", "Max Checks", "Max Hosts", "Minutes", "RTT", "Hop Count"]

        # Metics Header
        #header_list = ["Policy/Scanner/Scan", "Average per Asset in mins"]

        # Write the header to the csv
        agent_writer.writerow(header_list)

        # This function is used to parse the data
        # Getting the length of the Category and using it to get the average
        def average_by_policy(name, scan_info):
            # Print the Category to the Screen ( Scanner, Policy, Scan Name)
            print("\n{:100s} {:25s} {:10}".format(name, "AVG Minutes Per/Asset", "Total Assets"))
            print("-" * 20)

            # Write the category to the csv for delination
            #agent_writer.writerow([name])

            # Cycle through each category
            for scan in scan_info.items():

                # data is in a list [asset_uuid, mins] We need the length of the total mins found
                length = len(scan[1])

                # Reset the total per Category Item - Specific Scan ID, Scanner, Policy ID
                total = 0

                # Cycle through each asset record
                for assets in scan[1].values():
                    # Gather a total
                    total = assets + total

                # After calculating the total, lets get an average
                average = total/length

                # Print results to the screen
                print("\n{:100s} {:25d} {:10d}".format(scan[0], int(average), length))

                update = [scan[0], average]

                # Send the results as a list to the CSV
                #agent_writer.writerow(update)
            print("-" * 150)

        # Loop through each plugin 19506 and Parse data from it
        for vulns in plugin_data:

            # Output is the second item in the tuple from the DB
            plugin_output = vulns[1]

            # split the output by return
            parsed_output = plugin_output.split("\n")

            # grab the length so we can grab the seconds
            plugin_length = len(parsed_output)

            # grab the scan duration- second to the last variable
            duration = parsed_output[plugin_length - 2]

            # Split at the colon to grab the numerical value
            seconds = duration.split(" : ")

            # split to remove "secs"
            number = seconds[1].split(" ")

            # grab the number for our minute calculation
            final_number = number[0]

            if final_number != 'unknown':
                # convert seconds into minutes
                minutes = int(final_number) / 60

                # Grab data pair and split it at the colon and grab the values
                scan_name = parsed_output[9].split(" : ")[1]
                scan_policy = parsed_output[10].split(" : ")[1]
                scanner_ip = parsed_output[11].split(" : ")[1]
                scan_time = parsed_output[plugin_length - 3].split(" : ")[1]
                max_hosts = parsed_output[plugin_length- 8].split(" : ")[1]
                max_checks = parsed_output[plugin_length - 7].split(" : ")[1]

                if "no" not in parsed_output[14].split(" : ")[1]:
                    rtt = parsed_output[14].split(" : ")[1]
                else:
                    rtt = parsed_output[12].split(" : ")[1]

                try:
                    # Grab the last line in the Trace route Plugin output
                    # Split on the space and grab the numerical value.
                    hopcount = grab_hop_count(vulns[0])[0][-1].split(" ")[-1]
                except IndexError:
                    hopcount = "Unknown"

                parsed_data_organized = [vulns[0], scan_name, scan_policy, scanner_ip, scan_time, max_checks, max_hosts, minutes, rtt, hopcount]

                agent_writer.writerow(parsed_data_organized)

                # Organize Data by Scan Policy
                # If the category is not in the new dict, add it; else update it.
                if scan_policy not in scan_policy_dict:
                    scan_policy_dict[scan_policy] = {vulns[0]: minutes}
                else:
                    scan_policy_dict[scan_policy].update({vulns[0]: minutes})

                if scanner_ip not in scanner_dict:
                    scanner_dict[scanner_ip] = {vulns[0]: minutes}
                else:
                    scanner_dict[scanner_ip].update({vulns[0]: minutes})

                if scan_name not in scan_name_dict:
                    scan_name_dict[scan_name] = {vulns[0]: minutes}
                else:
                    scan_name_dict[scan_name].update({vulns[0]: minutes})

        average_by_policy("Policies", scan_policy_dict)

        average_by_policy("Scanners", scanner_dict)

        average_by_policy("Scan Name", scan_name_dict)



import click
from .database import new_db_connection, drop_tables
from .dbconfig import create_passwords_table


@click.command(help="Enter a ssh service account User Name and Password")
@click.option("--username", prompt=True, help="Provide your Access Key")
@click.option("--password", prompt=True, hide_input=True, help="Provide your Secret Key")
def ssh(username, password):
    # create all Tables when keys are added.
    database = r"navi.db"
    drop_conn = new_db_connection(database)
    drop_tables(drop_conn, 'ssh')

    create_passwords_table()

    ssh_dict = (username, password)
    database = r"navi.db"
    conn = new_db_connection(database)

    with conn:
        sql = '''INSERT or IGNORE into ssh(username, password) VALUES(?,?)'''
        cur = conn.cursor()
        cur.execute(sql, ssh_dict)



import click
from .database import new_db_connection
from .send_mail import send_attachment


def grab_smtp():
    # grab SMTP information
    database = r"navi.db"
    conn = new_db_connection(database)
    with conn:
        cur = conn.cursor()
        cur.execute("SELECT * from smtp;")
        rows = cur.fetchall()

        for row in rows:

            server = row[0]
            port = row[1]
            from_email = row[2]
            password = row[3]

        return server, port, from_email, password


@click.command(help="Mail yourself a Report")
@click.option("--message", default='', help="Email a custom message or use result of a navi command")
@click.option("--to", default='', help="Email address to send to")
@click.option("--subject", default='', help="Subject of the Email")
@click.option("-v", is_flag=True, help="Display a copy of the message")
@click.option("--file", default=None, help="The name of the file you want to attach to the email")
def mail(message, to, subject, v, file):
    try:
        # grab SMTP information
        server, port, from_email, password = grab_smtp()
        if to == '':
            to = input("Please enter the email you wish send this mail to: ")
        if subject == '':
            subject = input("Please enter the Subject of the email : ")

        subject += " - Emailed by navi"

        if v:
            click.echo("Here is a copy of your email that was Sent")
            click.echo(message)

        send_attachment(from_email, to, server, password, port, file, message, subject)

    except Exception as E:
        click.echo("Your Email information may be incorrect")
        click.echo("Run the 'SMTP' command to correct your information")
        click.echo(E)



import click
from .agent_export import agent_export
from .consec_export import consec_export
from .lumin_export import lumin_export
from .database import db_query
from .tag_export import tag_export
from .tag_helper import tag_checker
from .query_export import query_export
from .was_v2_export import was_export
from .agent_group_export import agent_group_export
from .user_export import user_export
from .compliance_export_csv import compliance_export_csv


@click.group(help="Export Tenable.io Data")
def export():
    pass


@export.command(help="Export All Asset data in the Navi Database to a CSV")
@click.option('--file', default="asset_data", help="Name of the file excluding 'csv'")
def assets(file):
    click.echo("\nExporting your data now. Saving {}.csv now...\n".format(file))
    asset_query = "select * from assets;"
    query_export(asset_query, file)


@export.command(help="Export All Agent data into a CSV")
def agents():
    click.echo("\nExporting your data now. Saving agent_data.csv now...\n")
    agent_export()


@export.command(help="Export Container Security Summary into a CSV")
def consec():
    click.echo("\nExporting your data now. Saving consec_data.csv now...\n")
    consec_export()


@export.command(help="Export Licensed Assets into a CSV")
@click.option('--file', default="licensed_data", help="Name of the file excluding 'csv'")
def licensed(file):
    click.echo("\nExporting your data now. Saving {}.csv now...\n".format(file))
    licensed_query = "SELECT ip_address, fqdn, uuid, last_licensed_scan_date from assets where last_licensed_scan_date != ' ';"
    query_export(licensed_query, file)


@export.command(help="Export All Asset data including ACR and AES into a CSV")
@click.option("-v", is_flag=True, help="Include ACR Drivers.  This will make a call per asset!")
@click.option('--file', default="lumin_data", help="Name of the file excluding 'csv'")
def lumin(v, file):

    if v:
        click.echo("\nExporting your data now. This could take some time.  300 Assets per minute max.")
        click.echo("Saving {}.csv now...\n".format(file))
        lumin_export()
    else:
        click.echo("\nExporting your data now.")
        click.echo("Saving {}.csv now...\n".format(file))
        asset_query = "select * from assets;"
        query_export(asset_query, file)


@export.command(help="Export All assets of a given network")
@click.argument('network_uuid')
@click.option('--file', default="network_data", help="Name of the file excluding 'csv'")
def network(network_uuid, file):
    click.echo("\nExporting your data now. Saving {}.csv now...".format(file))
    network_query = "SELECT * from assets where network=='{}';".format(network_uuid)
    query_export(network_query, file)


@export.command(help='Export assets by query to the vuln db')
@click.argument('statement')
@click.option('--file', default="query_data", help="Name of the file excluding 'csv'")
def query(statement, file):
    query_export(statement, file)


@export.command(help='Export Agents by Group name - API limits 5000 Agents')
@click.argument('group_name')
def group(group_name):
    click.echo("\nExporting your data now.  Saving agent_group_data.csv now...")
    agent_group_export(group_name)


@export.command(help="Export All assets by tag; Include ACR and AES into a CSV")
@click.option('--c', default=None, required=True, help="Export bytag with the following Category name")
@click.option('--v', default=None, required=True, help="Export bytag with the Tag Value; requires --c and Category Name")
@click.option('--ec', default=None, help="Exclude tag from export with Tag Category; requires --ev")
@click.option('--ev', default=None, help="Exclude tag from export with Tag Value; requires --ec")
@click.option('--file', default="bytag", help="Name of the file excluding 'csv'")
@click.option('-vulncounts', is_flag=True, help="Export Severity vulnerability Counts per asset. This will take some time...")
@click.option('--severity', type=click.Choice(['critical', 'high', 'medium', 'low', 'info'], case_sensitive=False), multiple=True)
def bytag(c, v, ec, ev, file, vulncounts, severity):
    new_list = []

    tag_assets = db_query("SELECT asset_uuid from tags where tag_key='" + c + "' and tag_value='" + v + "';")

    for asset in tag_assets:
        asset_uuid = asset[0]
        # Check the tag isn't apart of the exclude tag given
        if ev:
            check_for_no = tag_checker(asset_uuid, ec, ev)
            if check_for_no == 'no':
                new_list.append(asset_uuid)
        else:
            new_list.append(asset_uuid)

    if vulncounts:
        # Tell the export to pull verbose data - vunlcounts
        tag_export(new_list, file, 1)
    else:
        if severity:
            # If Severity is chosen then we will export vuln details
            if len(severity) == 1:
                # multiple choice values are returned as a tuple.
                # Here I break it out and put it in the format needed for sql
                asset_query = "select * from vulns where severity in ('{}');".format(severity[0])
                query_export(asset_query, file)
            else:
                # Here I just send the tuple in the query
                asset_query = "select * from vulns where severity in {};".format(severity)
                query_export(asset_query, file)

        else:
            # if Severity or vulncounts were not chosen we will export asset data from navi.db
            tag_export(new_list, file, 0)


@export.command(help="Export Webapp Scan Summary into a CSV - WAS V2")
def was():
    click.echo("\nExporting your data now. Saving was_summary_data.csv now...")
    was_export()


@export.command(help="Export User and Role information into a CSV")
def users():
    click.echo("\nExporting User Data now. Saving user-summary.csv now...\n")
    user_export()


@export.command(help="Export Compliance information into a CSV")
@click.option('--name', default=None, help="Exact name of the Audit file to be exported.  Use 'navi display audits' to "
                                           "get the right name")
@click.option('--uuid', default=None, help="UUID of the Asset for your export")
@click.option('--file', default="compliance_data", help="Name of the file excluding '.csv'")
def compliance(name, uuid, file):
    click.echo("\nExporting your requested Compliance data into a CSV\n")
    compliance_export_csv(name, uuid, file)


@export.command(help="Export All Vulnerability data in the Navi Database to a CSV")
@click.option('--file', default="vuln_data", help="Name of the file excluding '.csv'")
@click.option('--severity', type=click.Choice(['critical', 'high', 'medium', 'low', 'info'], case_sensitive=False), multiple=True)
def vulns(file, severity):
    click.echo("\nExporting your data now. Saving {}.csv now...\n".format(file))

    if severity:

        if len(severity) == 1:
            # multiple choice values are returned as a tuple.
            # Here I break it out and put it in the format needed for sql
            asset_query = "select * from vulns where severity in ('{}');".format(severity[0])
        else:
            # Here I just send the tuple in the query
            asset_query = "select * from vulns where severity in {};".format(severity)
    else:
        asset_query = "select * from vulns;"

    query_export(asset_query, file)



import click
from .api_wrapper import request_data


def get_attribute_uuid(name):
    attr_data = request_data('GET', '/api/v3/assets/attributes')
    uuid = 0
    for attr in attr_data['attributes']:
        if name == attr['name']:
            uuid = attr['id']

    return uuid


@click.group(help="Create and Assign Custom Attributes")
def attribute():
    pass


@attribute.command(help="Create an Custom Attribute")
@click.argument('name')
@click.option('--description', default='', help="Add a description for clarity")
def create(name, description):
    payload = {"attributes": [
        {
            "name": name,
            "description": "{} -Updated by Navi".format(description)
        }
    ]}
    data = request_data('POST', '/api/v3/assets/attributes', payload=payload)
    print(data)


@attribute.command(help="Add a custom attribute to an asset")
@click.option('--uuid', default='', help="UUID of the asset")
@click.option('--name', default='', help="Name of the Custom Attribute")
@click.option('--value', default='', help="Value of the Custom Attribute")
def assign(uuid, name, value):
    attr_uuid = get_attribute_uuid(name)
    print(attr_uuid)
    payload = {"attributes": [
        {
            "value": value,
            "id": attr_uuid
        }
    ]}
    assign_attr = request_data("PUT", '/api/v3/assets/{}/attributes'.format(uuid), payload=payload)
    print(assign_attr)



from .ip import ip
from .find import find
from .cs import cs
from .display import display
from .mac import mac
from .delete import delete
from .api import api
from .export import export
from .lumin import lumin
from .update import update
from .add import add
from .scan import scan
from .http import http
from .listen import listen
from .keys import keys
from .smtp import smtp
from .mail import mail
from .tag import tag
from .agroup import agroup
from .was import was
from .network import network
from .user import user
from .usergroup import usergroup
from .target_group import tgroup
from .cancel import cancel
from .migrate import migrate
from .agents import agent
from .push import push
from .ssh import ssh
from .scan_evaluation import evaluate
from .access import access
from .attribute import attribute


def plugin_loader(group):
    group.add_command(ip)
    group.add_command(find)
    group.add_command(cs)
    group.add_command(display)
    group.add_command(mac)
    group.add_command(delete)
    group.add_command(api)
    group.add_command(export)
    group.add_command(lumin)
    group.add_command(update)
    group.add_command(add)
    group.add_command(scan)
    group.add_command(http)
    group.add_command(listen)
    group.add_command(keys)
    group.add_command(smtp)
    group.add_command(mail)
    group.add_command(tag)
    group.add_command(agroup)
    group.add_command(was)
    group.add_command(network)
    group.add_command(user)
    group.add_command(usergroup)
    group.add_command(tgroup)
    group.add_command(cancel)
    group.add_command(migrate)
    group.add_command(agent)
    group.add_command(push)
    group.add_command(ssh)
    group.add_command(evaluate)
    group.add_command(access)
    group.add_command(attribute)




import csv
from .api_wrapper import request_data


def consec_export():
    data = request_data('GET', '/container-security/api/v2/images?limit=1000')
    with open('consec_data.csv', mode='w', encoding='utf-8') as csv_file:
        agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

        header_list = ["Container Name", "Docker ID", "# of Vulns"]
        agent_writer.writerow(header_list)

        for images in data["items"]:
            name = images["name"]
            docker_id = images["imageHash"]
            vulns = images["numberOfVulns"]
            agent_writer.writerow([name, docker_id, vulns])



import click
from .database import db_query
from .api_wrapper import request_data, tenb_connection

tio = tenb_connection()


def check_agroup_exists(aname):
    rvalue = 'no'
    for group in tio.access_groups.list():
        if str(group['name']).lower() == str(aname).lower():
            rvalue = group['id']
    return rvalue


@click.command(help="Create an Access group Based on a Tag - DEPRECATED in T.io")
@click.option('--name', default='', required=True, help="Choose a Name for your Access Group.")
@click.option('--c', default='', required=True, help="Tag Category name to use")
@click.option('--v', default='', required=True, help="Tag Value to use")
@click.option('--user', default='', help="The user you want to Assign to the Access Group - username@domain")
@click.option('--usergroup', default='', help="The User Group you want to assign to the Access Group")
@click.option('--perm', type=click.Choice(['scan', 'view', 'scanview'], case_sensitive=False), required=True)
def agroup(name, c, v, user, usergroup, perm):
    permission = []
    choice = 'none'
    permtype = 'none'

    if user == '' and usergroup == '':
        click.echo("\nYou Need to use '--user' or '--usergroup' command and supply "
                   "a user or group. e.g: user@yourdomain or Linux Admins\n")
        exit()

    if user != '':
        permtype = 'user'
        choice = user

    if usergroup != '':
        permtype = 'group'
        choice = usergroup

    if perm.lower() == 'scanview':
        permission = ["CAN_VIEW", "CAN_SCAN"]

    elif perm.lower() == 'view':
        permission = ["CAN_VIEW"]

    elif perm.lower() == 'scan':
        permission = ["CAN_SCAN"]

    assets = db_query("SELECT tag_uuid from tags where tag_key='" + c + "' and tag_value='" + v + "';")

    # Grab the first UUID...UUIDs returned are duplicates due to the db structure
    tag_uuid = [assets[0][0]]

    if tag_uuid:
        payload = {"name": str(name), "access_group_type": "MANAGE_ASSETS", "rules": [{"type": "tag_uuid", "operator": "set-has", "terms": tag_uuid}],
                   "principals": [{"permissions": permission, "type": permtype, "principal_name": choice}]}

        # Check to see if the group exists
        answer = check_agroup_exists(str(name))

        try:
            # Check to see if the list has any IPs in it.
            if answer == 'no':
                new_access_group = request_data('POST', '/v2/access-groups', payload=payload)
                click.echo("\nYour Access group {} is being created now \n".format(new_access_group['name']))
                click.echo("The UUID is {} \n".format(new_access_group['id']))
            else:
                update_group = request_data('PUT', '/v2/access-groups/' + str(answer), payload=payload)
                click.echo("\nYour Access group {} is being updated now \n".format(update_group['name']))
                click.echo("The UUID is {} \n".format(update_group['id']))
        except TypeError as E:
            click.echo("\nAccess group? - Check the Username")
            click.echo(E)
    else:
        click.echo("\nYour Tag was null. Check the spelling or perform a 'navi update assets'\n")



import pexpect
import click
from .database import db_query


PROMPT = ['#', '>>> ', '> ', '\$ ']


def send_command(shell, cmd, quick):
    if quick == 1:
        shell.sendline(cmd)
        shell.expect(PROMPT)
    else:
        shell.sendline(cmd)
        shell.expect(PROMPT, timeout=300)
        shell.expect(pexpect.TIMEOUT)

    response = shell.before
    print()
    print(response.decode('utf-8'))
    print()


def connect(user, host, password):
    ssh_new_key_string = 'Are you sure you want to continue connecting'
    ssh_login_string = 'ssh {}@{}'.format(user, host)

    shell = pexpect.spawn(ssh_login_string)

    return_code = shell.expect([pexpect.TIMEOUT, ssh_new_key_string, '[P|p]assword:'])

    if return_code == 0:
        print("error connecting")
        exit()

    if return_code == 1:
        shell.sendline('yes')
        second_return_code = shell.expect([pexpect.TIMEOUT, '[P|p]assword:'])

        if second_return_code == 0:
            print("error connecting")
            return

    shell.sendline(password)
    shell.expect(PROMPT)
    return shell


def scp(user, host, password, filename):
    ssh_new_key_string = 'Are you sure you want to continue connecting'
    scp_login_string = 'scp {} {}@{}:/'.format(filename, user, host)

    shell = pexpect.spawn(scp_login_string, timeout=300)

    return_code = shell.expect([pexpect.TIMEOUT, ssh_new_key_string, '[P|p]assword:'])

    if return_code == 0:
        print("error connecting")
        exit()

    if return_code == 1:
        shell.sendline('yes')
        second_return_code = shell.expect([pexpect.TIMEOUT, '[P|p]assword:'])

        if second_return_code == 0:
            print("error connecting")
            return

    shell.sendline(password)
    shell.expect(pexpect.EOF)
    response = shell.before
    print(response.decode('utf-8'))
    print()


@click.command(help="Push a command to a linux target")
@click.option('--command', default='', required=True, help="Command you want to run in double-quotes")
@click.option('--target', default='', required=True, help="Target IP receiving the command")
@click.option('--wait', is_flag=True, help="Wait for longer commands")
@click.option('--file', default=None, help="Push a file to a target")
def push(command, target, wait, file):

    try:
        credentials = db_query("select username, password from ssh;")
        user = credentials[0][0]
        password = credentials[0][1]
        if file:
            scp(user, target, password, file)
        else:
            shell = connect(user, target, password)

            if wait:
                send_command(shell, command, quick=0)
            else:
                send_command(shell, command, quick=1)

    except Exception as E:
        click.echo("Please use the 'navi ssh' command to enter your ssh credentials\n "
                   "If you have, then this host may not be a Linux machine or your Credentials are not working\n"
                   "Here is the Error: {}".format(E))



import click
from .plugins import plugin_loader


@click.group(help="This tool is not Supported by Tenable. \n Issues? Please submit an issue on Github.")
@click.pass_context
def cli(ctx):
    pass


plugin_loader(cli)



import click
from .api_wrapper import tenb_connection
import textwrap

tio = tenb_connection()


@click.group(help="Perform common tasks against Agents and Agent Groups")
def agent():
    pass


@agent.command(help="Display Agent information - Agent ID/UUID")
@click.option("--aid", default=None, help="Display Agent information using the Agent ID")
def display(aid):
    try:
        if aid:
            try:
                agent_details = tio.agents.details(aid)

                click.echo("\nAgent Details")
                click.echo("-----------------\n")
                click.echo("Agent Name: {}".format(agent_details['distro']))
                click.echo("Agent IP: {}".format(agent_details['distro']))
                click.echo("Agent UUID: {}".format(agent_details['uuid']))
                click.echo("Network UUID: {}".format(agent_details['network_uuid']))
                click.echo("Plugin Feed: {}".format(agent_details['plugin_feed_id']))

                click.echo("\nDistribution Information")
                click.echo("----------------------------\n")
                click.echo("Platform: {}".format(agent_details['platform']))
                click.echo("Distribution: {}".format(agent_details['distro']))
                click.echo("Core Version: {}".format(agent_details['core_version']))
                click.echo("Core Build: {}".format(agent_details['core_build']))

                click.echo("\nAgent Connection information")
                click.echo("----------------------------\n")
                click.echo("Last Connect Time: {}".format(agent_details['last_connect']))
                click.echo("Last Scan Time: {}".format(agent_details['last_scanned']))
                click.echo("Restart Pending: {}".format(agent_details['restart_pending']))
                click.echo("Status: {}".format(agent_details['status']))

                click.echo("\nAgent Groups")
                click.echo("----------------------------\n")
                for agent_groups in agent_details['groups']:
                    click.echo("Group Name({}): {}".format(str(agent_groups['id']), str(agent_groups['name'])))
            except TypeError:
                click.echo("\nCommon...really? Focus...You need the Agent ID... Try again\n")
                exit()

        else:
            click.echo("\n*** To see Agent Details use: navi agent --aid <agent id> ***\n")
            click.echo("\n{:45s} {:12} {}".format("Agent Name", "Agent ID", "Group(id)s"))
            click.echo("-" * 150)

            for agent_info in tio.agents.list():
                groups_string = ''
                try:
                    for group in agent_info['groups']:
                        groups_string = groups_string + ", {}({})".format(group['name'], group['id'])
                except KeyError:
                    pass
                click.echo("{:45s} {:12} {}".format(textwrap.shorten(str(agent_info['name']), width=45),
                                                    str(agent_info['id']),
                                                    textwrap.shorten(groups_string[1:], width=90)))
        click.echo()
    except AttributeError:
        click.echo("\nCheck your permissions or your API keys\n")


@agent.command(help="Display Agent Groups and membership information ")
@click.option("--gid", default=None, help="Display the agents that are members of the group using the group ID")
def groups(gid):

    if gid:
        group_details = tio.agent_groups.details(gid)
        click.echo("\n{:85s} {:15} {:40}".format("Agent Name", "Agent ID", "UUID", "Status"))
        click.echo("-" * 150)

        for agent_info in group_details['agents']:
            click.echo("{:85s} {:15} {:40s}".format(textwrap.shorten(str(agent_info['name']), width=85),
                                                    str(agent_info['id']),
                                                    str(agent_info['uuid']), str(agent_info['status'])))

        click.echo()
    else:
        click.echo("\n*** To see group membership use: navi agent groups --gid <group id> ***\n")
        try:
            click.echo("\n{:45} {:40} {:10}".format("Group Name", "Group UUID", "Group ID"))
            click.echo("-" * 150)

            group_list = tio.agent_groups.list()

            for group in group_list:
                click.echo("{:45} {:40} {:10}".format(str(group['name']), str(group['uuid']), str(group['id'])))

            click.echo()
        except AttributeError:
            click.echo("\nCheck your permissions or your API keys\n")


@agent.command(help="Create a new Agent Group")
@click.option("--name", default=None, required=True, help="The name of the new Agent Group you want to create")
@click.option("--scanner", default=1, help="Add Agent Group to a specific scanner")
def create(name, scanner):
    try:
        group_creation = tio.agent_groups.create(name=name, scanner_id=scanner)
    
        click.echo("\nYour agent group: {} has been created.\n\nHere is the ID: {} and UUID : {}"
                   "\n".format(group_creation['name'], str(group_creation['id']), str(group_creation['uuid'])))
    except AttributeError:
        click.echo("Check your API Keys")


@agent.command(help="Add an agent to a Group")
@click.option("--aid", default=None, required=True, help="The agent ID of the agent you want to add ")
@click.option("--gid", default=None, required=True, help="The Group ID you want to add the agent(s) to.")
@click.option("--file", default=None, required=False, help="CSV with UUIDs as the first column.")
def add(aid, gid, file):

    if file:
        # ignore AID and use the file instead
        for agent_info in tio.agents.list():

            print(agent_info['uuid'], agent_info['id'])

        import csv
        with open(file, 'r', newline='') as new_file:
            agent_list = []
            add_agents = csv.reader(new_file)

            for rows in add_agents:
                # UUID will be in the first column
                agent_list.append(rows[0])

            for agent_info in tio.agents.list():
                agent_uuid = agent_info['uuid']
                agent_id = agent_info['id']
                if agent_uuid in agent_list:
                    # Add agents to the Group
                    tio.agent_groups.add_agent(gid, agent_id)
    else:
        # Add agent to Group
        tio.agent_groups.add_agent(gid, aid)

        click.echo("\nYour agent has been added to the Group ID: {}\n".format(gid))


@agent.command(help="Remove an Agent from a Agent Group")
@click.option("--aid", default=None, required=True, help="The agent ID of the agent you want to remove ")
@click.option("--gid", default=None, required=True, help="The Group ID you want to add the agent(s) to.")
def remove(aid, gid):
    try:
        # Remove an agent from a Group
        tio.agent_groups.delete_agent(gid, aid)

        click.echo("\nYour agent has been removed from the Group ID: {}\n".format(gid))
    except AttributeError:
        click.echo("Check your API Keys")


@agent.command(help="Unlink an by Agent ID")
@click.option("--aid", default=None, required=True, help="The Agent ID of the agent you want to unlink")
def unlink(aid):
    try:
        tio.agents.unlink(aid)
        click.echo("\nYour Agent: {} has been unlinked".format(aid))
    except AttributeError:
        click.echo("Check your API Keys")



import csv
import click
from sqlite3 import Error
from .database import new_db_connection


def query_export(query, name):
    database = r"navi.db"
    conn = new_db_connection(database)
    with conn:
        cur = conn.cursor()
        try:
            cur.execute(query)
        except Error:
            print("\n No data! \n Please run 'navi update' first. OR check your query.\n")

        # Grab the Data
        data = cur.fetchall()

        # Grab the Table data
        descripts = cur.description

        # Create Header list
        header = []

        # Pull the row data out to popluate the headers
        for desc in descripts:
            header.append(desc[0])

        click.echo("I'm exporting {}.csv with your requested data".format(name))

        # Crete a csv file object

        with open('{}.csv'.format(name), mode='w', encoding='utf-8') as csv_file:
            agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')
            agent_writer.writerow(header)
            # Loop through each asset
            for assets in data:
                agent_writer.writerow(assets)



import click
import csv
from .database import new_db_connection, db_query
from .api_wrapper import request_data, tenb_connection
from .tag_helper import update_tag, confirm_tag_exists, return_tag_uuid, grab_all_tags, remove_tag
from sqlite3 import Error


def tag_by_tag(c, v, d, cv, cc, match):
    # BUG: Does not check to see if rule exists right now.

    tag_uuid = 0
    # Start a blank rules list to store current a new tag rule.
    rules_list = []

    # Does the Parent Tag exist?
    parent_answer = confirm_tag_exists(c, v)

    # Is this the parent tag new or current?
    if parent_answer == 'yes':

        # Does the Child Tag Exist?
        child_answer = confirm_tag_exists(cc, cv)

        # Is the child tag new or current?
        if child_answer == 'yes':

            # Update the tag parent tag with the new child tag
            click.echo("Your tag is being updated\n")

            try:
                # Need to grab the Tag UUID of our Parent Tag so we can get more details
                tag_data = request_data('GET', '/tags/values')
                for value in tag_data['values']:
                    if value['category_name'] == str(c):
                        if value['value'] == str(v):
                            try:
                                tag_uuid = value['uuid']

                                # Get filter details
                                current_rule_set = request_data("GET", "/tags/values/" + tag_uuid)

                                # The filter is a string in the API, pull out the dictionary representation and
                                filter_string = current_rule_set['filters']['asset']

                                # Turn the string into a dictionary
                                rule_set_dict = eval(filter_string)

                                # Identify 'or' vs 'and' and set the current filter list to our 'rule_list'
                                try:
                                    rules_list = rule_set_dict['and']
                                    match = 'and'
                                except KeyError:
                                    rules_list = rule_set_dict['or']
                                    match = 'or'

                                rules_list.append({"field": "tag.{}".format(cc), "operator": "set-has", "value": str(cv)})
                            except Exception as F:
                                click.echo(F)

                payload = {"category_name": str(c), "value": str(v), "description": str(d), "filters": {"asset": {str(match): rules_list}}}
                # Update the Parent Tag with the new child tag information
                data = request_data('PUT', '/tags/values/' + tag_uuid, payload=payload)

                value_uuid = data["uuid"]
                cat_uuid = data['category_uuid']
                click.echo("\nI've Updated your Tag - {} : {}\n".format(c, v))
                click.echo("The Category UUID is : {}\n".format(cat_uuid))
                click.echo("The Value UUID is : {}\n".format(value_uuid))
            except Exception as E:
                click.echo(E)
        else:
            click.echo("The Child Tag does not exist")

    else:
        # If the parent tag doesn't exist, does the child?
        # Check to see if the child tag exists
        child_answer = confirm_tag_exists(cc, cv)
        if child_answer == 'yes':
            # if the child tag does exist, then create the new tag with the existing tag as a child
            try:
                payload = {"category_name": str(c), "value": str(v), "description": str(d), "filters":
                           {"asset": {str(match): [{"field": "tag.{}".format(cc), "operator": "set-has", "value": str(cv)}]}}}
                data = request_data('POST', '/tags/values', payload=payload)

                value_uuid = data["uuid"]
                cat_uuid = data['category_uuid']
                click.echo("\nI've created your new Tag - {} : {}\n".format(c, v))
                click.echo("The Category UUID is : {}\n".format(cat_uuid))
                click.echo("The Value UUID is : {}\n".format(value_uuid))

            except Exception as F:
                click.echo(F)
        else:
            click.echo("Your Child Tag doesn't exist.\n You need to create a Child tag before adding it to a parent")


def tag_by_ip(ip_list, tag_list, c, v, d):
    # Tagging by IP is limited to 2000 Assets and is only used by the file command
    try:
        payload = {"category_name": str(c), "value": str(v), "description": str(d), "filters":
                   {"asset": {"and": [{"field": "ipv4", "operator": "eq", "value": str(ip_list[1:])}]}}}
        data = request_data('POST', '/tags/values', payload=payload)
        try:
            value_uuid = data["uuid"]
            cat_uuid = data['category_uuid']
            click.echo("\nI've created your new Tag - {} : {}\n".format(c, v))
            click.echo("The Category UUID is : {}\n".format(cat_uuid))
            click.echo("The Value UUID is : {}\n".format(value_uuid))
            click.echo("{} IPs added to the Tag".format(str(len(tag_list))))
        except Exception as E:
            click.echo("Duplicate Tag Category: You may need to delete your tag first\n")
            click.echo("We could not confirm your tag name, is it named weird?\n")
            click.echo(E)
    except:
        click.echo("Duplicate Category")


def tag_by_tenable_uuid(tag_list, c, v, d):
    # Tagging by Tenable UUID is limited to 500 UUIDS
    # Two problems here.  One: There is no PUT equivalant to chunk tenable UUIDs into groups of 500
    # two: the UUID returned from the agent endpoints is not the one needed to add a tag to the asset.
    try:
        payload = {"category_name": str(c), "value": str(v), "description": str(d), "filters":
                   {"asset": {"and": [{"field": "tenable_uuid", "operator": "eq", "value": str(tag_list[1:])}]}}}
        data = request_data('POST', '/tags/values', payload=payload)
        try:
            value_uuid = data["uuid"]
            cat_uuid = data['category_uuid']
            click.echo("\nI've created your new Tag - {} : {}\n".format(c, v))
            click.echo("The Category UUID is : {}\n".format(cat_uuid))
            click.echo("The Value UUID is : {}\n".format(value_uuid))
        except Exception as E:
            click.echo("Duplicate Tag Category: You may need to delete your tag first\n")
            click.echo("We could not confirm your tag name, is it named weird?\n")
            click.echo(E)
    except:
        click.echo("Duplicate Category")


def tag_by_uuid(tag_list, c, v, d):

    # Generator to split IPs into 2000 IP chunks
    def chunks(l, n):
        for i in range(0, len(l), n):
            yield l[i:i + n]

    # We Want to bail if the result is 0 Assets
    if not tag_list:
        click.echo("\nYour tag resulted in 0 Assets, therefore the tag wasn't created\n")
        exit()
    else:
        # Before updating confirm if the tag exists
        answer = confirm_tag_exists(c, v)

        # If the Tag does exist, update it by UUID.
        if answer == 'yes':
            # Check to see if the List of UUIDs is over 1999 (API Limit)
            if len(tag_list) > 1999:
                # break the list into 2000 IP chunks
                for chunks in chunks(tag_list, 1999):
                    update_tag(c, v, chunks)
            else:
                # If the Chunk is less than 2000, simply update it.
                update_tag(c, v, tag_list)
        # If the tag doesn't exist. we need to create one.
        else:
            # Create the Tag
            payload = {"category_name": str(c), "value": str(v), "description": str(d)}
            data = request_data('POST', '/tags/values', payload=payload)
            value_uuid = data["uuid"]
            cat_uuid = data['category_uuid']
            click.echo("\nI've created your new Tag - {} : {}\n".format(c, v))
            click.echo("The Category UUID is : {}\n".format(cat_uuid))
            click.echo("The Value UUID is : {}\n".format(value_uuid))

            # Check to see if the List of UUIDs is over 1999 (API Limit)
            if len(tag_list) > 1999:
                try:
                    click.echo("Your Tag list was over 2000 IPs.  Splitting the UUIDs into chunks and updating the tags now")
                    # Break the UUIDs into Chunks and update the tag per chunk
                    for chunks in chunks(tag_list, 1999):
                        update_tag(c, v, chunks)

                except Exception as E:
                    click.echo("An Error Occurred: \n")
                    click.echo(E)
            else:
                try:
                    update_tag(c, v, tag_list)
                except Exception as E:
                    click.echo("An Error Occurred: \n")
                    click.echo(E)


def download_csv_by_plugin_id(scan_id, hist_id):
    filename = f'{scan_id}-report.csv'
    tio = tenb_connection()

    # Stream the report to disk
    with open(filename, 'wb') as fobj:
        tio.scans.export(scan_id, ('plugin.id', 'eq', '19506'),
                         format='csv', fobj=fobj, history_id=hist_id)
    return filename


def create_uuid_list(filename):
    from csv import DictReader
    uuids = []
    with open(filename) as fobj:
        for row in DictReader(fobj):
            asset_uuid = row['Asset UUID']
            if asset_uuid and asset_uuid != '':
                uuids.append(asset_uuid)
    return uuids


def remove_uuids_from_tag(tag_uuid):
    # Create an list to store our asset uuids
    asset_uuid_list = []

    tag_data = db_query("select asset_uuid from assets LEFT JOIN tags ON uuid == asset_uuid where tag_uuid=='{}';".format(str(tag_uuid)))

    # Clean up the data into a list
    for asset_uuid_loop in tag_data:
        asset_uuid_list.append(asset_uuid_loop[0])

    # Generator to split IPs into 2000 IP chunks
    def chunks(l, n):
        for i in range(0, len(l), n):
            yield l[i:i + n]

    # Check to see if the List of UUIDs is over 1999 (API Limit)
    if len(asset_uuid_list) > 1999:

        # break the list into 2000 IP chunks
        for chunks in chunks(asset_uuid_list, 1999):
            remove_tag(str(tag_uuid), chunks)
    else:
        # If the Chunk is less than 2000, simply update it.
        remove_tag(str(tag_uuid), asset_uuid_list)


@click.command(help="Create a Tag Category/Value Pair")
@click.option('--c', default='', help="Create a Tag with the following Category name")
@click.option('--v', default='', help="Create a Tag Value; requires --c and Category Name or UUID")
@click.option('--d', default='This Tag was created/updated by navi', help="Description for your Tag")
@click.option('--plugin', default='', help="Create a tag by plugin ID")
@click.option('--name', default='', help="Create a Tag by the text found in the Plugin Name")
@click.option('--group', default='', help="Create a Tag based on a Agent Group")
@click.option('--output', default='', help="Create a Tag based on the text in the output. Requires --plugin")
@click.option('--port', default='', help="Create a Tag based on Assets that have a port open.")
@click.option('--file', default='', help="Create a Tag based on IPs in a CSV file.")
@click.option('--scantime', default='', help="Create a Tag for assets that took longer than supplied minutes")
@click.option('--cc', default='', help="Add a Tag to a new parent tag: Child Category")
@click.option('--cv', default='', help="Add a Tag to a new parent tag: Child Value")
@click.option('--scanid', default='', help="Create a tag by Scan ID")
@click.option('--pipe', default='', help="Create a Tag based on a pipe from a 'navi find query -pipe' command")
@click.option('-all', is_flag=True, help="Change Default Match rule of 'or' to 'and'")
@click.option('--query', default='', help="Use a custom query to create a tag.")
@click.option('--remove', default='', help="Remove this tag from all assets to support ephemeral asset tagging")
def tag(c, v, d, plugin, name, group, output, port, scantime, file, cc, cv, scanid, pipe, all, query, remove):
    # start a blank list
    tag_list = []
    ip_list = ""

    if c == '':
        click.echo("Category is required.  Please use the --c command")
        exit()

    if v == '':
        click.echo("Value is required. Please use the --v command")
        exit()

    if output != '' and plugin == '':
        click.echo("You must supply a plugin")
        exit()

    if plugin:
        try:
            database = r"navi.db"
            conn = new_db_connection(database)
            with conn:
                cur = conn.cursor()
                # See if we want to refine our search by the output found in this plugin
                # this needs to have a JOIN statement to reduce the amount
                if output != "":
                    cur.execute("SELECT asset_ip, asset_uuid, output from vulns where plugin_id='" + plugin + "' and output LIKE '%" + output + "%';")
                else:
                    cur.execute("SELECT asset_ip, asset_uuid, output from vulns where plugin_id=%s;" % plugin)

                plugin_data = cur.fetchall()
                for x in plugin_data:
                    ip = x[0]
                    uuid = x[1]
                    # To reduce duplicates check for the UUID in the list.
                    if uuid not in tag_list:
                        tag_list.append(uuid)
                        ip_list = ip_list + "," + ip
                    else:
                        pass
        except Error:
            pass

        tag_by_uuid(tag_list, c, v, d)

    if port != '':
        database = r"navi.db"
        conn = new_db_connection(database)
        with conn:
            cur = conn.cursor()
            cur.execute("SELECT asset_uuid from vulns where port=" + port + " and (plugin_id='11219' or plugin_id='14272' or plugin_id='14274' or plugin_id='34220' or plugin_id='10335');")

            data = cur.fetchall()

            try:
                for vulns in data:
                    uuid = vulns[0]
                    # To reduce duplicates check for the UUID in the list.
                    if uuid not in tag_list:
                        tag_list.append(uuid)
            except ValueError:
                pass
        tag_by_uuid(tag_list, c, v, d)

    if name != '':
        try:
            database = r"navi.db"
            conn = new_db_connection(database)
            with conn:
                cur = conn.cursor()
                cur.execute("SELECT asset_ip, asset_uuid, output from vulns where plugin_name LIKE '%" + name + "%';")

                plugin_data = cur.fetchall()
                for x in plugin_data:
                    ip = x[0]
                    uuid = x[1]
                    if uuid not in tag_list:
                        tag_list.append(uuid)
                        ip_list = ip_list + "," + ip
                    else:
                        pass
        except Error:
            pass

        tag_by_uuid(tag_list, c, v, d)

    if group != '':
        from uuid import UUID

        try:
            offset = 0
            total = 0
            while offset <= total:
                querystring = {"limit": "5000", "offset": offset}
                group_data = request_data('GET', '/scanners/1/agent-groups', params=querystring)

                for agent_group in group_data['groups']:
                    group_name = agent_group['name']
                    group_id = agent_group['id']

                    # Match on Given Name
                    if group_name == group:
                        agent_data = request_data('GET', '/scanners/1/agent-groups/' + str(group_id) + '/agents', params=querystring)
                        total = agent_data['pagination']['total']

                        for agent in agent_data['agents']:
                            new_uuid = UUID(agent['uuid']).hex
                            tag_uuid = db_query("select uuid from assets where agent_uuid='{}'".format(new_uuid))
                            # New agents will not have a corresponding UUID and will throw an error.
                            # Since they can not be tagged without a UUID, we will silently skip over the agent
                            try:
                                tag_list.append(tag_uuid[0][0])
                            except IndexError:
                                pass
                offset = offset + 5000
        except Error:
            click.echo("You might not have agent groups, or you are using Nessus Manager.  ")

        tag_by_uuid(tag_list, c, v, d)

    if scantime != '':
        database = r"navi.db"
        conn = new_db_connection(database)
        with conn:
            cur = conn.cursor()
            cur.execute("SELECT asset_ip, asset_uuid, output from vulns where plugin_id='19506';")

            data = cur.fetchall()
            try:
                for vulns in data:

                    output = vulns[2]

                    # split the output by return
                    parsed_output = output.split("\n")

                    # grab the length so we can grab the seconds
                    length = len(parsed_output)

                    # grab the scan duration- second to the last variable
                    duration = parsed_output[length - 2]

                    # Split at the colon to grab the numerical value
                    seconds = duration.split(" : ")

                    # split to remove "secs"
                    number = seconds[1].split(" ")

                    # grab the number for our minute calculation
                    final_number = number[0]

                    # convert seconds into minutes
                    minutes = int(final_number) / 60

                    # grab assets that match the criteria
                    if minutes > int(scantime):
                        try:
                            ip_list = ip_list + "," + str(vulns[0])
                            tag_list.append(vulns[1])
                        except ValueError:
                            pass
                click.echo()
            except ValueError:
                pass

        tag_by_uuid(tag_list, c, v, d)

    if file != '':
        with open(file, 'r', newline='') as new_file:
            add_ips = csv.reader(new_file)

            for row in add_ips:
                for ips in row:
                    # need to look grab UUIDS per IP for the ablity to update Tags
                    tag_list.append(ips)
                    ip_list = ip_list + "," + ips

        tag_by_ip(ip_list, tag_list, c, v, d)

    if cv != '' and cc != '':
        if all:
            match = 'and'
            tag_by_tag(c, v, d, cv, cc, match)
        else:
            match = 'or'
            tag_by_tag(c, v, d, cv, cc, match)

    if scanid:
        tag_list = []
        try:
            scandata = request_data('GET', '/scans/' + str(scanid))

            status = scandata['history'][0]['status']

            if status == 'completed':
                try:
                    for host in scandata['hosts']:
                        tag_list.append(host['uuid'])

                    # This method is to get around the 5000 row return limit in IO.
                    if len(tag_list) >= 4999:

                        click.echo("\nYou're scan is 5000 IPs or More. Downloading, Parsing and Cleaning up scans to ensure all assets are tagged\n")
                        click.echo("\nTags can take a few minutes to populate in the UI when applied to 1000s of assets\n")
                        hist_id = scandata['history'][0]['history_id']
                        filename = download_csv_by_plugin_id(scanid, hist_id)
                        tag_list = create_uuid_list(filename)
                        tag_by_uuid(tag_list, c, v, d)

                        import os
                        os.remove(filename)
                    else:
                        # Scans under 5000 would just use the current tag_list
                        tag_by_uuid(tag_list, c, v, d)
                except TypeError:
                    click.echo("Check the scan ID")
                except KeyError:
                    click.echo("The scan used is archived, canceled, imported or aborted. Your Tag was not created.")
            else:
                # cycle trough history until you reach completed
                hist = 1
                new_scan_data = None
                new_hist = 0

                # if you got here the first scan has not completed
                while status != 'completed':
                    # call the scan history again to check the next Status
                    # continue to do so until one says completed
                    new_scan_data = request_data('GET', '/scans/' + str(scanid))
                    status = new_scan_data['history'][hist]['status']

                    # Grab the history of the completed asset
                    new_hist = new_scan_data['history'][hist]['history_id']

                    # To limit the amount of called to a minimum
                    # Use a while loop and increase the history ID by 1 until completed is found
                    hist = hist + 1

                try:
                    # To reduce API calls lets just download the csv
                    filename = download_csv_by_plugin_id(scanid, new_hist)
                    tag_list = create_uuid_list(filename)
                    tag_by_uuid(tag_list, c, v, d)

                    import os
                    os.remove(filename)

                except TypeError:
                    click.echo("Check the scan ID")
                except KeyError:
                    click.echo("The scan used is archived, canceled, imported or aborted. Your Tag was not created.")

        except IndexError:
            click.echo("\nThe scan used is archived, canceled, imported or aborted and has no completed scans in it's history\n")
        except Exception as E:
            click.echo("Check your Scan ID; An Error occurred\n{}".format(E))

    if pipe:
        tag_by_uuid(eval(pipe), c, v, d)

    if query:

        if ("uuid" or "asset_uuid") in str(query):

            data = db_query(query)

            for asset in set(data):
                tag_list.append(asset[0])

            tag_by_uuid(tag_list, c, v, d)

        else:
            click.echo("\nYou have to return uuids or asset_uuid. \nYour query must have uuid or asset_uuid\n")
            exit()

    if remove != '':

        if remove == 'byname':

            # Find the UUID of our given tag
            tag_list = grab_all_tags()
            try:
                for tag_info in tag_list:
                    # Grab our UUID
                    if str(tag_info[0]).lower() == str(c).lower():
                        if str(tag_info[1]).lower() == str(v).lower():
                            new_uuid = str(tag_info[2])
                            print("grab all tags, find correct uuid, hit db with {}".format(new_uuid))
                            remove_uuids_from_tag(new_uuid)

            except Exception as E:
                click.echo(E)

        else:
            remove_uuids_from_tag(remove)



import csv
import click
from .database import db_query
from .api_wrapper import tenb_connection

tio = tenb_connection()


def tag_export(tag_list, filename, option):
    click.echo("This export can take some time.  ~300 assets per minute.")

    # Create our headers - We will Add these two our list in order
    if option == 1:
        header_list = ["IP Address", "Hostname", "FQDN", "UUID", "First Found", "Last Found", "Operating System",
                       "Mac Address", "Agent-UUID", "last Licensed Scan Date", 'Network', 'ACR', 'AES', 'AWS ID',
                       'Info', 'Low', 'Medium', 'High', 'Critical']
    else:
        header_list = ["IP Address", "Hostname", "FQDN", "UUID", "First Found", "Last Found", "Operating System",
                       "Mac Address", "Agent-UUID", "last Licensed Scan Date", 'Network', 'ACR', 'AES', 'AWS ID']

    # Crete a csv file object
    with open('{}.csv'.format(filename), mode='w', encoding='utf-8') as csv_file:
        agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

        # write our Header information first
        agent_writer.writerow(header_list)

        for uuid in tag_list:
            export_list = []
            data = db_query("SELECT * from assets where uuid='{}';".format(uuid))

            # SQL gives us a tuple, we need to convert it to a list.
            for item in data[0]:
                export_list.append(item)

            if option == 1:
                try:
                    asset_info = tio.workbenches.asset_info(uuid)

                    for vuln in asset_info['counts']['vulnerabilities']['severities']:
                        export_list.append(vuln["count"])  # Add the vuln counts to the new list

                except ConnectionError:
                    click.echo("Check your API keys or your internet connection")

            # write to the CSV
            agent_writer.writerow(export_list)
            print("\n{}".format(export_list))
        click.echo("\nExport success! - {}.csv\n".format(filename))



import click
from .api_wrapper import request_data
from .error_msg import error_msg


@click.group(help="Check container Compliance or get a container vuln report")
def cs():
    pass


@cs.command(help="Report CVSS 7 or Above")
@click.option('--container', default='', help='Report CVSS 7 or above by \'/repository/image/tag\'')
@click.option('--docker', default='', help='Report CVSS 7 or above by Docker ID')
def report(container, docker):
    if container:
        try:
            data = request_data('GET', '/container-security/api/v2/reports' + str(container))
            try:
                for vulns in data['findings']:
                    if float(vulns['nvdFinding']['cvss_score']) >= 7:
                        click.echo("CVE ID : {}".format(vulns['nvdFinding']['cve']))
                        click.echo("CVSS Score : {}".format(vulns['nvdFinding']['cvss_score']))
                        click.echo("-" * 22)
                        click.echo("\nDescription : \n\n {}".format(vulns['nvdFinding']['description']))
                        click.echo("\nRemediation : \n\n {}".format(vulns['nvdFinding']['remediation']))
                        click.echo("----------------------END-------------------------\n")
            except TypeError:
                click.echo("This Container has no data or is not found")
            except ValueError:
                pass
        except Exception as E:
            error_msg(E)

    if docker:
        try:
            data = request_data('GET', '/container-security/api/v1/reports/by_image?image_id=' + str(docker))

            try:
                for vulns in data['findings']:
                    if float(vulns['nvdFinding']['cvss_score']) >= 7:
                        click.echo("CVE ID : {}".format(vulns['nvdFinding']['cve']))
                        click.echo("CVSS Score : {}".format(vulns['nvdFinding']['cvss_score']))
                        click.echo("-" * 22)
                        click.echo("\nDescription : \n\n {}".format(vulns['nvdFinding']['description']))
                        click.echo("\nRemediation : \n\n {}".format(vulns['nvdFinding']['remediation']))
                        click.echo("----------------------END-------------------------\n")
            except TypeError:
                click.echo("This Container has no data or is not found")
            except ValueError:
                pass
        except Exception as E:
            error_msg(E)


@cs.command(help="Verify if a image passes compliance")
@click.argument('image')
def comply(image):
    try:
        data = request_data('GET', '/container-security/api/v1/policycompliance?image_id=' + str(image))

        click.echo("Status : {}".format(data['status']))
    except Exception as E:
        error_msg(E)



import click
import textwrap
from sqlite3 import Error
from .api_wrapper import tenb_connection, request_data
from .database import db_query


def plugin_by_ip(ipaddr, plugin):
    try:
        if len(ipaddr) < 17:
            rows = db_query("SELECT output, cves, score, state from vulns where asset_ip=\"%s\" and plugin_id=%s" % (ipaddr, plugin))
        else:
            rows = db_query("SELECT output, cves, score, state from vulns where asset_uuid=\"%s\" and plugin_id=%s" % (ipaddr, plugin))

        for plug in rows:
            click.echo("\nCurrent Plugin State: {} ".format(plug[3]))
            if plug[2] != ' ':
                click.echo("\nVPR Score: {}".format(plug[2]))

            click.echo("\nPlugin Output")
            click.echo("-" * 60)
            click.echo(plug[0])

            if plug[1] != ' ':
                click.echo("CVEs attached to this plugin")
                click.echo("-" * 80)
                click.echo("{}\n".format(plug[1]))
        click.echo()
    except IndexError:
        click.echo("No information found for this plugin")


def vulns_by_uuid(uuid):
    try:
        data = db_query("select plugin_id, plugin_name, plugin_family, port, protocol, severity, state from vulns where asset_uuid='{}' and severity !='info';".format(uuid))

        click.echo("\n{:10s} {:70s} {:35s} {:10s} {:6s} {:6s} {}".format("Plugin", "Plugin Name", "Plugin Family", "state", "Port", "Proto", "Severity"))
        click.echo("-"*150)

        for vulns in data:
            plugin_id = vulns[0]
            plugin_name = vulns[1]
            plugin_family = vulns[2]
            port = vulns[3]
            protocol = vulns[4]
            severity = vulns[5]
            state = vulns[6]
            click.echo("{:10s} {:70s} {:35s} {:10s} {:6s} {:6s} {}".format(plugin_id, textwrap.shorten(plugin_name, 70), textwrap.shorten(plugin_family, 35), state, port, protocol, severity))
        click.echo("")
    except Error as e:
        click.echo(e)


def get_attributes(uuid):
    attr_data = request_data('GET', '/api/v3/assets/{}/attributes'.format(uuid))
    return attr_data


def info_by_uuid(uuid):
    try:
        data = db_query("select plugin_id, plugin_name, plugin_family, port, protocol, severity from vulns where asset_uuid='{}' and severity =='info';".format(uuid))

        click.echo("\n{:10s} {:90s} {:25s} {:6s} {:6s} {}".format("Plugin", "Plugin Name", "Plugin Family", "Port", "Proto", "Severity"))
        click.echo("-"*150)

        for vulns in data:
            plugin_id = vulns[0]
            plugin_name = vulns[1]
            plugin_family = vulns[2]
            port = vulns[3]
            protocol = vulns[4]
            severity = vulns[5]
            click.echo("{:10s} {:90s} {:25s} {:6s} {:6s} {}".format(plugin_id, plugin_name, plugin_family, port, protocol, severity))
        click.echo("")
    except Error as e:
        click.echo(e)


def cves_by_uuid(uuid):
    try:
        data = db_query("select plugin_id, cves from vulns where asset_uuid='{}' and cves !=' ';".format(uuid))

        click.echo("\n{:10s} {}".format("Plugin", "CVEs"))
        click.echo("-"*150)

        for vulns in data:
            plugin_id = vulns[0]
            cves = vulns[1]
            click.echo("{:10s} {}".format(plugin_id, textwrap.shorten(cves, 140)))
        click.echo("")
    except IndexError:
        click.echo("Something went wrong")


@click.command(help="Get Asset details based on IP or UUID")
@click.argument('ipaddr')
@click.option('--plugin', default='', help='Find Details on a particular plugin ID')
@click.option('-n', is_flag=True, help='Netstat Established(58561) and Listening and Open Ports(14272)')
@click.option('-p', is_flag=True, help='Patch Information - 66334')
@click.option('-t', is_flag=True, help='Trace Route - 10287')
@click.option('-o', is_flag=True, help='Process Information - 70329')
@click.option('-c', is_flag=True, help='Connection Information - 64582')
@click.option('-s', is_flag=True, help='Services Running - 22964')
@click.option('-r', is_flag=True, help='Local Firewall Rules - 56310')
@click.option('-patches', is_flag=True, help='Missing Patches - 38153')
@click.option('-d', is_flag=True, help="Scan Detail: 19506 plugin output")
@click.option('-software', is_flag=True, help="Find software installed on Unix(22869) of windows(20811) hosts")
@click.option('-outbound', is_flag=True, help="outbound connections found by nnm")
@click.option('-exploit', is_flag=True, help="Display Solution, Description for each Exploit")
@click.option('-critical', is_flag=True, help="Display Plugin Output for each Critical Vuln")
@click.option('-details', is_flag=True, help="Details on an Asset: IP, UUID, Vulns, etc")
@click.option('-vulns', is_flag=True, help="Display all vulnerabilities and their plugin IDs")
@click.option('-info', is_flag=True, help="Display all info plugins and their IDs")
@click.option('-cves', is_flag=True, help="Display all cves found on the asset")
@click.option('-compliance', is_flag=True, help="Display all Compliance info for a given asset UUID")
@click.pass_context
def ip(ctx, ipaddr, plugin, n, p, t, o, c, s, r, patches, d, software, outbound, exploit, critical, details, vulns,
       info, cves, compliance):
    tio = tenb_connection()

    if d:
        click.echo('\nScan Detail')
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(19506))

    if n:
        click.echo("\nNetstat info")
        click.echo("Established and Listening")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(58651))
        click.echo("\nNetstat Open Ports")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(14272))

    if p:
        click.echo("\nPatch Information")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(66334))

    if t:
        click.echo("\nTrace Route Info")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(10287))

    if o:
        click.echo("\nProcess Info")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(70329))
        plugin_by_ip(ipaddr, str(110483))

    if patches:
        click.echo("\nMissing Patches")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(38153))
        plugin_by_ip(ipaddr, str(66334))

        click.echo("\nLast Reboot")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(56468))

    if c:
        click.echo("\nConnection info")
        click.echo("-" * 15)
        click.echo()
        plugin_by_ip(ipaddr, str(64582))

    if s:
        try:
            if len(ipaddr) < 17:
                data = db_query("SELECT output, port from vulns where asset_ip=\"%s\" and plugin_id='22964'" % ipaddr)
            else:
                data = db_query("SELECT output, port from vulns where asset_uuid=\"%s\" and plugin_id='22964'" % ipaddr)

            for plugins in data:
                output = plugins[0]
                port = plugins[1]
                click.echo("\n{} {}".format(str(output), str(port)))
            click.echo()
        except IndexError:
            click.echo("No information for plugin 22964")

    if r:
        click.echo("Local Firewall Info")
        click.echo("-" * 15)
        plugin_by_ip(ipaddr, str(56310))
        plugin_by_ip(ipaddr, str(61797))

    if software:
        try:
            plugin_by_ip(ipaddr, str(22869))
            plugin_by_ip(ipaddr, str(20811))
        except IndexError:
            click.echo("No Software found")

    if outbound:
        try:
            if len(ipaddr) < 17:
                data = db_query("SELECT output, port, protocol from vulns where asset_ip=\"%s\" and plugin_id='16'" % ipaddr)
            else:
                data = db_query("SELECT output, port, protocol from vulns where asset_uuid=\"%s\" and plugin_id='16'" % ipaddr)

            click.echo("\n{:15s} {:5} {}".format("IP address", "Port", "Protocol"))
            click.echo("-" * 25)
            for plugins in data:
                output = plugins[0]
                port = plugins[1]
                proto = plugins[2]
                click.echo("\n{:15s} {:5} {}".format(str(output), str(port), str(proto)))
            click.echo()
        except Exception as E:
            click.echo("No information for plugin 16")
            click.echo(E)

    if exploit:
        try:
            if len(ipaddr) < 17:
                data = db_query("SELECT uuid from assets where ip_address='" + ipaddr + "';")
            else:
                data = ipaddr

            for assets in data:
                asset_id = assets[0]

                click.echo("\nExploitable Details for : {}\n".format(ipaddr))

                vuln_data = tio.workbenches.asset_vulns(asset_id, ("plugin.attributes.exploit_available", "eq", "true"), age=90)

                for plugins in vuln_data:
                    plugin = plugins['plugin_id']

                    plugin_data = tio.plugins.plugin_details(plugin)

                    click.echo("\n----Exploit Info----")
                    click.echo(plugin_data['name'])
                    click.echo()
                    for attribute in plugin_data['attributes']:

                        if attribute['attribute_name'] == 'cve':
                            cve = attribute['attribute_value']
                            click.echo("CVE ID : " + cve)

                        if attribute['attribute_name'] == 'description':
                            description = attribute['attribute_value']
                            click.echo("Description")
                            click.echo("------------\n")
                            click.echo(description)
                            click.echo()

                        if attribute['attribute_name'] == 'solution':
                            solution = attribute['attribute_value']
                            click.echo("\nSolution")
                            click.echo("------------\n")
                            click.echo(solution)
                            click.echo()
        except Exception as E:
            click.echo(E)

    if critical:
        try:
            if len(ipaddr) < 17:

                data = db_query("SELECT uuid from assets where ip_address='" + ipaddr + "';")
            else:
                data = db_query("SELECT uuid from assets where uuid='" + ipaddr + "';")

            for assets in data:
                asset_id = assets[0]
                click.echo("\nCritical Vulns for Ip Address : {}\n".format(ipaddr))

                asset_vulns = tio.workbenches.asset_vulns(asset_id, age=90)

                for severities in asset_vulns:
                    vuln_name = severities["plugin_name"]
                    plugin_id = severities["plugin_id"]
                    severity = severities["severity"]
                    state = severities["vulnerability_state"]

                    # only pull the critical vulns; critical = severity 4
                    if severity >= 4:
                        click.echo("Plugin Name : {}".format(vuln_name))
                        click.echo("ID : {}".format(str(plugin_id)))
                        click.echo("Severity : Critical")
                        click.echo("State : {}".format(state))
                        click.echo("----------------\n")
                        plugin_by_ip(str(ipaddr), str(plugin_id))
                        click.echo()
        except Exception as E:
            click.echo(E)

    if details:
        if len(ipaddr) < 17:
            data = db_query("SELECT uuid from assets where ip_address='" + ipaddr + "';")
        else:
            data = db_query("SELECT uuid from assets where uuid='" + ipaddr + "';")

        for assets in data:
            asset_data = tio.workbenches.asset_info(str(assets[0]))

            try:
                asset_id = asset_data['id']

                click.echo("\nTenable ID")
                click.echo("--------------")
                click.echo(asset_id)

                click.echo("\nNetwork Name")
                click.echo("--------------")
                click.echo(asset_data['network_name'])

                click.echo("\nIdentities")
                click.echo("--------------")
                try:
                    for netbioss in asset_data['netbios_name']:
                        click.echo("Netbios - {}".format(netbioss))
                except KeyError:
                    pass
                try:
                    for fqdns in asset_data['fqdns']:
                        click.echo("FQDN - {}".format(fqdns))
                except KeyError:
                    pass

                try:
                    for hosts in asset_data['hostname']:
                        click.echo("Host Name - {}".format(hosts))
                except KeyError:
                    pass

                try:
                    for agentname in asset_data['agent_name']:
                        click.echo("Agent Name - {}".format(agentname))
                except KeyError:
                    pass

                try:
                    for awsid in asset_data['aws_ec2_instance_id']:
                        click.echo("AWS EC2 Instance ID - {}".format(awsid))
                except KeyError:
                    pass

                try:
                    for awsamiid in asset_data['aws_ec2_ami_id']:
                        click.echo("AWS EC2 AMI ID - {}".format(awsamiid))
                except KeyError:
                    pass

                try:
                    for awsname in asset_data['aws_ec2_name']:
                        click.echo("AWS EC2 Name - {}".format(awsname))
                except KeyError:
                    pass

                click.echo("\nOperating Systems")
                click.echo("--------------")
                try:
                    for oss in asset_data['operating_system']:
                        click.echo(oss)
                except KeyError:
                    pass

                try:
                    click.echo("\nIP Addresses:")
                    click.echo("--------------")
                    for ips in asset_data['ipv4']:
                        click.echo(ips)
                except KeyError:
                    pass

                try:
                    click.echo("\nMac Addresses:")
                    click.echo("--------------")
                    for macs in asset_data['mac_address']:
                        click.echo(macs)
                except KeyError:
                    pass

                try:
                    click.echo("\nCloud Information:")
                    click.echo("--------------")
                    for zone in asset_data['aws_availability_zone']:
                        click.echo("AWS Availability Zone - {}".format(zone))
                except KeyError:
                    pass

                try:
                    for groupname in asset_data['aws_ec2_instance_group_name']:
                        click.echo("AWS Instance group Name - {}".format(groupname))
                except KeyError:
                    pass

                try:
                    for zone in asset_data['aws_availability_zone']:
                        click.echo("AWS Availability Zone - {}".format(zone))
                except KeyError:
                    pass
                try:
                    for statename in asset_data['aws_ec2_instance_state_name']:
                        click.echo("AWS Instance State - {}".format(statename))
                except KeyError:
                    pass
                try:
                    for instatncetype in asset_data['aws_ec2_instance_type']:
                        click.echo("AWS Instance Type - {}".format(instatncetype))
                except KeyError:
                    pass
                try:
                    for region in asset_data['aws_region']:
                        click.echo("AWS Region - {}".format(region))
                except KeyError:
                    pass

                try:
                    for subnet in asset_data['aws_subnet_id']:
                        click.echo("AWS Subnet ID - {}".format(subnet))
                except KeyError:
                    pass
                try:
                    for vpc in asset_data['aws_vpc_id']:
                        click.echo("AWS VPC ID - {}".format(vpc))
                except KeyError:
                    pass
                try:
                    for azureid in asset_data['azure_resource_id']:
                        click.echo("Azure Resource ID - {}".format(azureid))
                except KeyError:
                    pass
                try:
                    for vmid in asset_data['azure_vm_id']:
                        click.echo("Azure VM ID - {}".format(vmid))
                except KeyError:
                    pass

                try:
                    for gcpid in asset_data['gcp_instance_id']:
                        click.echo("GCP Instance ID - {}".format(gcpid))
                except KeyError:
                    pass
                try:
                    for projectid in asset_data['gcp_project_id']:
                        click.echo("GCP Project ID- {}".format(projectid))
                except KeyError:
                    pass
                try:
                    for gcpzone in asset_data['gcp_zone']:
                        click.echo("GCP Zone - {}".format(gcpzone))
                except KeyError:
                    pass
                try:
                    click.echo("\nSources:")
                    click.echo("-" * 15)
                    for source in asset_data['sources']:
                        click.echo(source['name'])
                except KeyError:
                    pass
                try:
                    click.echo("\nTags:")
                    click.echo("-" * 15)
                    for tags in asset_data['tags']:
                        click.echo("{} : {}".format(tags["tag_key"], tags['tag_value']))
                except KeyError:
                    pass
                try:
                    click.echo("\nCustom Attributes:")
                    click.echo("-" * 15)
                    for attr in get_attributes(ipaddr)['attributes']:
                        click.echo("{} : {}".format(attr['name'], attr['value']))

                except KeyError:
                    pass
                click.echo("\nVulnerability Counts")
                click.echo("-" * 15)

                asset_info = tio.workbenches.asset_info(asset_id)

                for vuln in asset_info['counts']['vulnerabilities']['severities']:
                    click.echo("{} : {}".format(vuln["name"], vuln["count"]))

                try:
                    click.echo("\nAsset Exposure Score : {}".format(asset_info['exposure_score']))
                    click.echo("\nAsset Criticality Score : {}".format(asset_info['acr_score']))
                except KeyError:
                    pass

                click.echo("\nLast Authenticated Scan Date - {}".format(asset_data['last_authenticated_scan_date']))
                click.echo("\nLast Licensed Scan Date - {}".format(asset_data['last_licensed_scan_date']))
                click.echo("-" * 50)
                click.echo("-" * 50)
            except KeyError:
                pass

    if vulns:
        if len(ipaddr) < 17:
            data = db_query("SELECT uuid from assets where ip_address='" + ipaddr + "';")
            for assets in data:
                click.echo("\nAsset UUID: {}".format(ipaddr))
                click.echo("Asset IP: {}".format(ipaddr))
                click.echo("-" * 26)
                vulns_by_uuid(assets[0])
        else:
            data = db_query("SELECT ip_address from assets where uuid='" + ipaddr + "';")
            click.echo("\nAsset UUID: {}".format(ipaddr))
            click.echo("Asset IP: {}".format(data[0]))
            click.echo("-" * 26)
            vulns_by_uuid(ipaddr)

    if cves:
        if len(ipaddr) < 17:
            data = db_query("SELECT uuid from assets where ip_address='" + ipaddr + "';")

            for assets in data:
                click.echo("\nAsset UUID: {}".format(assets[0]))
                click.echo("Asset IP: {}".format(ipaddr))
                click.echo("-" * 26)
                cves_by_uuid(assets[0])
        else:
            data = db_query("SELECT ip_address from assets where uuid='" + ipaddr + "';")
            click.echo("\nAsset UUID: {}".format(ipaddr))
            click.echo("Asset IP: {}".format(data[0]))
            click.echo("-" * 26)
            cves_by_uuid(ipaddr)

    if info:
        if len(ipaddr) < 17:
            data = db_query("SELECT uuid from assets where ip_address='" + ipaddr + "';")

            for assets in data:
                click.echo("\nAsset UUID: {}".format(assets[0]))
                click.echo("Asset IP: {}".format(ipaddr))
                click.echo("-" * 26)
                info_by_uuid(assets[0])
        else:
            data = db_query("SELECT ip_address from assets where uuid='" + ipaddr + "';")
            click.echo("\nAsset UUID: {}".format(ipaddr))
            click.echo("Asset IP: {}".format(data[0]))
            click.echo("-" * 26)
            info_by_uuid(ipaddr)

    if plugin != '':
        plugin_by_ip(ipaddr, plugin)

    if compliance:
        if len(ipaddr) > 16:
            compliance_data = db_query("SELECT check_name, status, audit_file from compliance where asset_uuid='{}';".format(ipaddr))
            click.echo("{:84} {:8} {}".format("Check Name", "Status", "Audit File"))
            click.echo("-" * 150)
            for finding in compliance_data:
                check_name = finding[0]
                status = finding[1]
                audit_file = finding[2]
                click.echo("{:85} {:8} {}".format(textwrap.shorten(check_name, width=80), status,
                                                  textwrap.shorten(audit_file, width=60)))
        else:
            click.echo("\nCompliance info requires a UUID\n\nFor simplicity I pulled the UUID(s) with this IP\n")
            uuid_data = db_query("SELECT uuid, fqdn from assets where ip_address='{}';".format(ipaddr))

            click.echo("{:45}{}".format("UUID", "FQDN"))
            click.echo("-" * 150)
            for address in uuid_data:
                click.echo("{:45}{}".format(address[0], address[1]))
            click.echo()



import click
from .api_wrapper import tenb_connection

tio = tenb_connection()


def nessus_scanners():
    try:
        click.echo("\n{:35s} {:20} {}".format("Scanner Name", "Scanner ID", "Scanner UUID"))
        click.echo("-" * 150)
        for scanners in tio.scanners.list():
            click.echo("{:35s} {:20} {}".format(str(scanners["name"]), str(scanners["id"]), str(scanners['uuid'])))
        click.echo()
    except AttributeError:
        click.echo("\nCheck your permissions or your API keys\n")



import pprint
import click
from .api_wrapper import request_data
from .error_msg import error_msg


@click.command(help="Explore the API using simple GET requests ex: 'navi api /scans ")
@click.argument('url')
@click.option('-raw', is_flag=True, help="Return raw Json")
@click.option('--limit', default=50, help="Change API Request Limit")
@click.option('--offset', default=0, help="Change API Request Offset")
def api(url, raw, limit, offset):
    params = {"limit": limit, "offset": offset}
    try:
        data = request_data('GET', url, params=params)
        if not raw:
            pprint.pprint(data)
        else:
            click.echo(data)

    except Exception as E:
        error_msg(E)



import click
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders


def send_attachment(from_email, to_email, mail_server, password, port, filename, message, subject):
    try:
        server = smtplib.SMTP(mail_server, port)
        server.ehlo()
        server.starttls()
        server.login(from_email, password)

        msg2 = MIMEMultipart()
        msg2['Subject'] = subject
        msg2['From'] = from_email
        msg2['To'] = to_email
        msg2.attach(MIMEText(message, 'plain'))

        # If there is an attachment, add it to the email.
        if filename:
            attachment = open(filename, "rb")
            p = MIMEBase('application', 'octet-stream')

            # To change the payload into encoded form
            p.set_payload(attachment.read())

            # encode into base64
            encoders.encode_base64(p)

            p.add_header('Content-Disposition', "attachment; filename= %s" % filename)

            # attach the instance 'p' to instance 'msg'
            msg2.attach(p)

        server.sendmail(from_email, to_email, msg2.as_string())
        server.close()

        click.echo('Email sent!')
    except Exception as E:
        click.echo('\nSomething went wrong...Your email information may be incorrect\n')
        click.echo("For now you must use an 'less secure' setting in Gmail for this feature."
                   "\n I'm working on a more secure option.")
        click.echo(E)



import time
import threading
import click
from queue import Queue
from sqlite3 import Error
from .api_wrapper import request_data
from .database import new_db_connection, drop_tables, insert_vulns, insert_update_info, get_last_update_id, db_query
from .dbconfig import create_vulns_table

lock = threading.Lock()

q = Queue()


def worker():
    # The worker thread pulls an item from the queue and processes it
    while True:
        item = q.get()
        chunk_num = item[58:]
        # parse_data(request_data('GET', item))
        data = request_data('GET', item)
        parse_data(data, chunk_num)
        q.task_done()


def parse_data(chunk_data, chunk_number):
    database = r"navi.db"
    vuln_conn = new_db_connection(database)
    vuln_conn.execute('pragma journal_mode=wal;')
    vuln_conn.execute('pragma cashe_size=-10000')
    vuln_conn.execute('pragma synchronous=OFF')
    with vuln_conn:
        try:
            # loop through all of the vulns in this chunk
            for vulns in chunk_data:
                # create a blank list to append asset details
                vuln_list = []
                # Try block to ignore assets without IPs
                try:
                    try:
                        ipv4 = vulns['asset']['ipv4']
                        vuln_list.append(ipv4)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        asset_uuid = vulns['asset']['uuid']
                        vuln_list.append(asset_uuid)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        hostname = vulns['asset']['hostname']
                        vuln_list.append(hostname)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        first_found = vulns['first_found']
                        vuln_list.append(first_found)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        last_found = vulns['last_found']
                        vuln_list.append(last_found)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        output = vulns['output']
                        vuln_list.append(output)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        plugin_id = vulns['plugin']['id']
                        vuln_list.append(plugin_id)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        plugin_name = vulns['plugin']['name']
                        vuln_list.append(plugin_name)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        plugin_family = vulns['plugin']['family']
                        vuln_list.append(plugin_family)
                    except KeyError:
                        vuln_list.append(" ")
                    try:
                        port = vulns['port']['port']
                        vuln_list.append(port)
                    except KeyError:
                        vuln_list.append(" ")
                    try:
                        protocol = vulns['port']['protocol']
                        vuln_list.append(protocol)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        severity = vulns['severity']
                        vuln_list.append(severity)
                    except KeyError:
                        vuln_list.append(" ")
                    try:
                        scan_completed = vulns['scan']['completed_at']
                        vuln_list.append(scan_completed)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        scan_started = vulns['scan']['started_at']
                        vuln_list.append(scan_started)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        scan_uuid = vulns['scan']['uuid']
                        vuln_list.append(scan_uuid)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        schedule_id = vulns['scan']['schedule_id']
                        vuln_list.append(schedule_id)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        state = vulns['state']
                        vuln_list.append(state)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        cves = vulns['plugin']['cve']
                        vuln_list.append(str(cves))
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        score = vulns['plugin']['vpr']['score']
                        vuln_list.append(score)
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        exploit = vulns['plugin']['exploit_available']
                        vuln_list.append(str(exploit))
                    except KeyError:
                        vuln_list.append(" ")

                    try:
                        insert_vulns(vuln_conn, vuln_list)
                    except Error as e:
                        click.echo(e)
                except IndexError:
                    click.echo("skipped one")
        except TypeError:
            click.echo("Your Export has no data.  It may have expired")
            click.echo("Error on Chunk {}".format(chunk_number))

    click.echo("Chunk {} Finished".format(chunk_number))
    vuln_conn.close()


def vuln_export(days, ex_uuid, threads):
    start = time.time()

    database = r"navi.db"
    drop_conn = new_db_connection(database)
    drop_conn.execute('pragma journal_mode=wal;')

    # Right now we just drop the table.  Eventually I will actually update the database
    drop_tables(drop_conn, 'vulns')

    create_vulns_table()

    # Set URLS for threading
    urls = []

    # Set the payload to the maximum number of assets to be pulled at once
    day = 86400
    new_limit = day * int(days)
    day_limit = time.time() - new_limit
    pay_load = {"num_assets": 50, "filters": {'last_found': int(day_limit), "state": ['open', 'reopened']}}
    try:

        if ex_uuid == '0':
            # request an export of the data
            export = request_data('POST', '/vulns/export', payload=pay_load)

            # grab the export UUID
            ex_uuid = export['export_uuid']
            click.echo('\nRequesting Vulnerability Export with ID : {}'.format(ex_uuid))

            # set a variable to True for our While loop
            not_ready = True
        else:
            click.echo("\nUsing your Export UUID\n")
            not_ready = True

        # now check the status
        status = request_data('GET', '/vulns/export/' + ex_uuid + '/status')

        # status = get_data('/vulns/export/89ac18d9-d6bc-4cef-9615-2d138f1ff6d2/status')
        click.echo("Status : {}".format(str(status["status"])))

        # loop to check status until finished
        while not_ready is True:
            # Pull the status, then pause 5 seconds and ask again.
            if status['status'] == 'PROCESSING' or 'QUEUED':
                time.sleep(2.5)
                status = request_data('GET', '/vulns/export/' + ex_uuid + '/status')
                # click.echo("Status : " + str(status["status"]))

            # Exit Loop once confirmed finished
            if status['status'] == 'FINISHED':
                ptime = time.time()
                click.echo("\nProcessing Time took : {}".format(str(ptime - start)))

                # Display how many chunks there are
                avail = status['total_chunks']
                click.echo("\nChunks Available - {}".format(avail))
                click.echo("Downloading chunks now...hold tight...This can take some time\n")
                not_ready = False

            # Tell the user an error occured
            if status['status'] == 'ERROR':
                click.echo("Error occurred")

        # grab all of the chunks and craft the URLS for threading
        for y in status['chunks_available']:
            urls.append('/vulns/export/' + ex_uuid + '/chunks/' + str(y))

        for i in range(threads):
            t = threading.Thread(target=worker)
            t.daemon = True  # thread dies when main thread (only non-daemon thread) exits.
            t.start()

        for item in range(len(urls)):
            q.put(urls[item])

        q.join()
        end = time.time()
        click.echo("Vulnerability Update Time took : {}\n".format(str(end - start)))

        update_id = get_last_update_id()
        diff_dict = [update_id, str(start), str(days), "Vuln update", str(ex_uuid)]  # need to ignore if the Ex_uuid exists in the db.
        database_2 = r"navi.db"
        conn = new_db_connection(database_2)
        with conn:
            insert_update_info(conn, diff_dict)

        click.echo("\nCreating a few indexes to make queries faster.\n")
        db_query("CREATE INDEX vulns_plugin_id on vulns (plugin_id);")
        db_query("CREATE INDEX vulns_ports on vulns (port);")
        db_query("CREATE INDEX vulns_uuid on vulns (asset_uuid);")

    except KeyError:
        click.echo("Well this is a bummer; you don't have permissions to download Asset data :( ")

    except TypeError:
        click.echo("You may not be authorized or your keys are invalid")
